from datetime import datetime
from typing import ClassVar, Dict, Optional, List, Type, cast

from attrs import define, field

from resoto_plugin_aws.resource.autoscaling import AwsAutoScalingGroup
from resoto_plugin_aws.resource.base import AwsResource, GraphBuilder, AwsApiSpec
from resoto_plugin_aws.resource.iam import AwsIamRole
from resotolib.baseresources import BaseAccount, EdgeType, ModelReference  # noqa: F401
from resotolib.json_bender import Bender, S, Bend, ForallBend
from resotolib.types import Json
from resoto_plugin_aws.aws_client import AwsClient


# todo: annotate with no serialization annotation
class EKSTaggable:
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="tag_resource",
                    result_name=None,
                    resourceArn=self.arn,
                    tags=[{key: value}],
                )
                return True
            return False
        return False

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        if isinstance(self, AwsResource):
            if spec := self.api_spec:
                client.call(
                    service=spec.service,
                    action="untag_resource",
                    result_name=None,
                    resourceArn=self.arn,
                    tagKeys=[key],
                )
                return True
            return False
        return False


@define(eq=False, slots=False)
class AwsEksNodegroupScalingConfig:
    kind: ClassVar[str] = "aws_eks_nodegroup_scaling_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "min_size": S("minSize"),
        "max_size": S("maxSize"),
        "desired_size": S("desiredSize"),
    }
    min_size: Optional[int] = field(default=None)
    max_size: Optional[int] = field(default=None)
    desired_size: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsEksRemoteAccessConfig:
    kind: ClassVar[str] = "aws_eks_remote_access_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "ec2_ssh_key": S("ec2SshKey"),
        "source_security_groups": S("sourceSecurityGroups", default=[]),
    }
    ec2_ssh_key: Optional[str] = field(default=None)
    source_security_groups: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsEksTaint:
    kind: ClassVar[str] = "aws_eks_taint"
    mapping: ClassVar[Dict[str, Bender]] = {"key": S("key"), "value": S("value"), "effect": S("effect")}
    key: Optional[str] = field(default=None)
    value: Optional[str] = field(default=None)
    effect: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsEksNodegroupResources:
    kind: ClassVar[str] = "aws_eks_nodegroup_resources"
    mapping: ClassVar[Dict[str, Bender]] = {
        "auto_scaling_groups": S("autoScalingGroups", default=[]) >> ForallBend(S("name")),
        "remote_access_security_group": S("remoteAccessSecurityGroup"),
    }
    auto_scaling_groups: List[str] = field(factory=list)
    remote_access_security_group: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsEksIssue:
    kind: ClassVar[str] = "aws_eks_issue"
    mapping: ClassVar[Dict[str, Bender]] = {
        "code": S("code"),
        "message": S("message"),
        "resource_ids": S("resourceIds", default=[]),
    }
    code: Optional[str] = field(default=None)
    message: Optional[str] = field(default=None)
    resource_ids: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsEksNodegroupHealth:
    kind: ClassVar[str] = "aws_eks_nodegroup_health"
    mapping: ClassVar[Dict[str, Bender]] = {"issues": S("issues", default=[]) >> ForallBend(AwsEksIssue.mapping)}
    issues: List[AwsEksIssue] = field(factory=list)


@define(eq=False, slots=False)
class AwsEksNodegroupUpdateConfig:
    kind: ClassVar[str] = "aws_eks_nodegroup_update_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "max_unavailable": S("maxUnavailable"),
        "max_unavailable_percentage": S("maxUnavailablePercentage"),
    }
    max_unavailable: Optional[int] = field(default=None)
    max_unavailable_percentage: Optional[int] = field(default=None)


@define(eq=False, slots=False)
class AwsEksLaunchTemplateSpecification:
    kind: ClassVar[str] = "aws_eks_launch_template_specification"
    mapping: ClassVar[Dict[str, Bender]] = {"name": S("name"), "version": S("version"), "id": S("id")}
    name: Optional[str] = field(default=None)
    version: Optional[str] = field(default=None)
    id: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsEksNodegroup(EKSTaggable, AwsResource):
    # Note: this resource is collected via AwsEksCluster
    kind: ClassVar[str] = "aws_eks_nodegroup"
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {"default": ["aws_eks_cluster"], "delete": ["aws_eks_cluster", "aws_autoscaling_group"]},
        "successors": {"default": ["aws_autoscaling_group"]},
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("nodegroupName"),
        "name": S("nodegroupName"),
        "tags": S("tags", default={}),
        "cluster_name": S("clusterName"),
        "ctime": S("createdAt"),
        "arn": S("nodegroupArn"),
        "version": S("version"),
        "group_release_version": S("releaseVersion"),
        "group_modified_at": S("modifiedAt"),
        "group_status": S("status"),
        "group_capacity_type": S("capacityType"),
        "group_scaling_config": S("scalingConfig") >> Bend(AwsEksNodegroupScalingConfig.mapping),
        "group_instance_types": S("instanceTypes", default=[]),
        "group_subnets": S("subnets", default=[]),
        "group_remote_access": S("remoteAccess") >> Bend(AwsEksRemoteAccessConfig.mapping),
        "group_ami_type": S("amiType"),
        "group_node_role": S("nodeRole"),
        "group_labels": S("labels"),
        "group_taints": S("taints", default=[]) >> ForallBend(AwsEksTaint.mapping),
        "group_resources": S("resources") >> Bend(AwsEksNodegroupResources.mapping),
        "group_disk_size": S("diskSize"),
        "group_health": S("health") >> Bend(AwsEksNodegroupHealth.mapping),
        "group_update_config": S("updateConfig") >> Bend(AwsEksNodegroupUpdateConfig.mapping),
        "group_launch_template": S("launchTemplate") >> Bend(AwsEksLaunchTemplateSpecification.mapping),
    }
    cluster_name: Optional[str] = field(default=None)
    group_nodegroup_arn: Optional[str] = field(default=None)
    group_version: Optional[str] = field(default=None)
    group_release_version: Optional[str] = field(default=None)
    group_modified_at: Optional[datetime] = field(default=None)
    group_status: Optional[str] = field(default=None)
    group_capacity_type: Optional[str] = field(default=None)
    group_scaling_config: Optional[AwsEksNodegroupScalingConfig] = field(default=None)
    group_instance_types: List[str] = field(factory=list)
    group_subnets: List[str] = field(factory=list)
    group_remote_access: Optional[AwsEksRemoteAccessConfig] = field(default=None)
    group_ami_type: Optional[str] = field(default=None)
    group_node_role: Optional[str] = field(default=None)
    group_labels: Optional[Dict[str, str]] = field(default=None)
    group_taints: List[AwsEksTaint] = field(factory=list)
    group_resources: Optional[AwsEksNodegroupResources] = field(default=None)
    group_disk_size: Optional[int] = field(default=None)
    group_health: Optional[AwsEksNodegroupHealth] = field(default=None)
    group_update_config: Optional[AwsEksNodegroupUpdateConfig] = field(default=None)
    group_launch_template: Optional[AwsEksLaunchTemplateSpecification] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if cluster_name := self.cluster_name:
            builder.dependant_node(
                self, clazz=AwsEksCluster, reverse=True, delete_same_as_default=True, name=cluster_name
            )
        if self.group_resources:
            for arn in self.group_resources.auto_scaling_groups:
                builder.dependant_node(self, clazz=AwsAutoScalingGroup, arn=arn)

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(
            service="eks",
            action="delete_nodegroup",
            result_name=None,
            clusterName=self.cluster_name,
            nodegroupName=self.name,
        )
        return True


@define(eq=False, slots=False)
class AwsEksVpcConfigResponse:
    kind: ClassVar[str] = "aws_eks_vpc_config_response"
    mapping: ClassVar[Dict[str, Bender]] = {
        "subnet_ids": S("subnetIds", default=[]),
        "security_group_ids": S("securityGroupIds", default=[]),
        "cluster_security_group_id": S("clusterSecurityGroupId"),
        "vpc_id": S("vpcId"),
        "endpoint_public_access": S("endpointPublicAccess"),
        "endpoint_private_access": S("endpointPrivateAccess"),
        "public_access_cidrs": S("publicAccessCidrs", default=[]),
    }
    subnet_ids: List[str] = field(factory=list)
    security_group_ids: List[str] = field(factory=list)
    cluster_security_group_id: Optional[str] = field(default=None)
    vpc_id: Optional[str] = field(default=None)
    endpoint_public_access: Optional[bool] = field(default=None)
    endpoint_private_access: Optional[bool] = field(default=None)
    public_access_cidrs: List[str] = field(factory=list)


@define(eq=False, slots=False)
class AwsEksKubernetesNetworkConfigResponse:
    kind: ClassVar[str] = "aws_eks_kubernetes_network_config_response"
    mapping: ClassVar[Dict[str, Bender]] = {
        "service_ipv4_cidr": S("serviceIpv4Cidr"),
        "service_ipv6_cidr": S("serviceIpv6Cidr"),
        "ip_family": S("ipFamily"),
    }
    service_ipv4_cidr: Optional[str] = field(default=None)
    service_ipv6_cidr: Optional[str] = field(default=None)
    ip_family: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsEksLogSetup:
    kind: ClassVar[str] = "aws_eks_log_setup"
    mapping: ClassVar[Dict[str, Bender]] = {"types": S("types", default=[]), "enabled": S("enabled")}
    types: List[str] = field(factory=list)
    enabled: Optional[bool] = field(default=None)


@define(eq=False, slots=False)
class AwsEksLogging:
    kind: ClassVar[str] = "aws_eks_logging"
    mapping: ClassVar[Dict[str, Bender]] = {
        "cluster_logging": S("clusterLogging", default=[]) >> ForallBend(AwsEksLogSetup.mapping)
    }
    cluster_logging: List[AwsEksLogSetup] = field(factory=list)


@define(eq=False, slots=False)
class AwsEksIdentity:
    kind: ClassVar[str] = "aws_eks_identity"
    mapping: ClassVar[Dict[str, Bender]] = {"oidc": S("oidc", "issuer")}
    oidc: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsEksEncryptionConfig:
    kind: ClassVar[str] = "aws_eks_encryption_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "resources": S("resources", default=[]),
        "provider": S("provider", "keyArn"),
    }
    resources: List[str] = field(factory=list)
    provider: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsEksConnectorConfig:
    kind: ClassVar[str] = "aws_eks_connector_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "activation_id": S("activationId"),
        "activation_code": S("activationCode"),
        "activation_expiry": S("activationExpiry"),
        "provider": S("provider"),
        "role_arn": S("roleArn"),
    }
    activation_id: Optional[str] = field(default=None)
    activation_code: Optional[str] = field(default=None)
    activation_expiry: Optional[datetime] = field(default=None)
    provider: Optional[str] = field(default=None)
    role_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsEksCluster(EKSTaggable, AwsResource):
    kind: ClassVar[str] = "aws_eks_cluster"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("eks", "list-clusters", "clusters")
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_iam_role"],
            "delete": [],
        }
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("name"),
        "tags": S("tags", default={}),
        "name": S("name"),
        "arn": S("arn"),
        "ctime": S("createdAt"),
        "cluster_version": S("version"),
        "cluster_endpoint": S("endpoint"),
        "cluster_role_arn": S("roleArn"),
        "cluster_resources_vpc_config": S("resourcesVpcConfig") >> Bend(AwsEksVpcConfigResponse.mapping),
        "cluster_kubernetes_network_config": S("kubernetesNetworkConfig")
        >> Bend(AwsEksKubernetesNetworkConfigResponse.mapping),
        "cluster_logging": S("logging") >> Bend(AwsEksLogging.mapping),
        "cluster_identity": S("identity") >> Bend(AwsEksIdentity.mapping),
        "cluster_status": S("status"),
        "cluster_certificate_authority": S("certificateAuthority", "data"),
        "cluster_client_request_token": S("clientRequestToken"),
        "cluster_platform_version": S("platformVersion"),
        "cluster_encryption_config": S("encryptionConfig", default=[]) >> ForallBend(AwsEksEncryptionConfig.mapping),
        "cluster_connector_config": S("connectorConfig") >> Bend(AwsEksConnectorConfig.mapping),
    }
    cluster_version: Optional[str] = field(default=None)
    cluster_endpoint: Optional[str] = field(default=None)
    cluster_role_arn: Optional[str] = field(default=None)
    cluster_resources_vpc_config: Optional[AwsEksVpcConfigResponse] = field(default=None)
    cluster_kubernetes_network_config: Optional[AwsEksKubernetesNetworkConfigResponse] = field(default=None)
    cluster_logging: Optional[AwsEksLogging] = field(default=None)
    cluster_identity: Optional[AwsEksIdentity] = field(default=None)
    cluster_status: Optional[str] = field(default=None)
    cluster_certificate_authority: Optional[str] = field(default=None)
    cluster_client_request_token: Optional[str] = field(default=None)
    cluster_platform_version: Optional[str] = field(default=None)
    cluster_encryption_config: List[AwsEksEncryptionConfig] = field(factory=list)
    cluster_connector_config: Optional[AwsEksConnectorConfig] = field(default=None)

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        for name in cast(List[str], json):
            cluster_json = builder.client.get("eks", "describe-cluster", "cluster", name=name)
            if cluster_json is not None:
                cluster = AwsEksCluster.from_api(cluster_json)
                builder.add_node(cluster, cluster_json)
                for ng_name in builder.client.list("eks", "list-nodegroups", "nodegroups", clusterName=name):
                    ng_json = builder.client.get(
                        "eks", "describe-nodegroup", "nodegroup", clusterName=name, nodegroupName=ng_name
                    )
                    if ng_json is not None:
                        ng = AwsEksNodegroup.from_api(ng_json)
                        builder.add_node(ng, ng_json)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        builder.add_edge(self, EdgeType.default, clazz=AwsIamRole, arn=self.cluster_role_arn)

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(service=self.api_spec.service, action="delete_cluster", result_name=None, name=self.name)
        return True


resources: List[Type[AwsResource]] = [AwsEksNodegroup, AwsEksCluster]



from resoto_plugin_aws.resource.iam import (
    AwsIamPolicy,
    AwsIamGroup,
    AwsIamServerCertificate,
    AwsIamRole,
    AwsIamUser,
    AwsIamAccessKey,
    AwsIamInstanceProfile,
)
from test.resources import round_trip_for
from typing import Any, cast
from types import SimpleNamespace
from resoto_plugin_aws.aws_client import AwsClient


def test_server_certificates() -> None:
    round_trip_for(AwsIamServerCertificate, "dns_names", "sha1_fingerprint")


def test_instance_profiles() -> None:
    round_trip_for(AwsIamInstanceProfile)


def test_user_roles_groups_policies_keys() -> None:
    _, builder = round_trip_for(AwsIamUser)

    # users ------------
    assert len(builder.resources_of(AwsIamUser)) == 3
    assert (test_user := builder.node(clazz=AwsIamUser, name="test_user")) is not None
    assert [p.policy_name for p in test_user.user_policies] == ["stsAssumeRole"]

    # keys ------------
    assert len(builder.resources_of(AwsIamAccessKey)) == 2
    assert (ak_test := builder.node(clazz=AwsIamAccessKey, id="ak_test")) is not None
    assert len(builder.graph.nodes) == 12
    # make sure access keys are created and connected as part of the user
    assert builder.graph.has_edge(test_user, ak_test)

    # groups ------------
    assert len(builder.resources_of(AwsIamGroup)) == 1
    assert (group := builder.node(clazz=AwsIamGroup, name="test_group")) is not None
    assert [p.policy_name for p in group.group_policies] == ["stsAssumeRole"]

    # policies ------------
    assert len(builder.resources_of(AwsIamPolicy)) == 4
    assert (policy := builder.node(clazz=AwsIamPolicy, name="master-elb")) is not None
    assert policy.policy_attachment_count == 1

    # roles
    assert len(builder.resources_of(AwsIamRole)) == 2
    assert (role := builder.node(clazz=AwsIamRole, name="role1")) is not None
    assert len(role.role_policies) == 1


def test_server_certificate_deletion() -> None:
    res, _ = round_trip_for(AwsIamServerCertificate, "dns_names", "sha1_fingerprint")

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_server_certificate"
        assert kwargs["ServerCertificateName"] == res.name

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.delete_resource(client)


def test_aws_iam_policy_deletion() -> None:
    _, builder = round_trip_for(AwsIamUser)
    res = builder.resources_of(AwsIamPolicy)[0]

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_policy"
        assert kwargs["PolicyArn"] == res.arn

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.delete_resource(client)


def test_aws_iam_group_deletion() -> None:
    _, builder = round_trip_for(AwsIamUser)
    res = builder.resources_of(AwsIamGroup)[0]

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_group"
        assert kwargs["GroupName"] == res.name

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.delete_resource(client)


def test_aws_iam_role_deletion() -> None:
    _, builder = round_trip_for(AwsIamUser)
    res = builder.resources_of(AwsIamRole)[0]

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_role"
        assert kwargs["RoleName"] == res.name

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.delete_resource(client)


def test_aws_iam_user_deletion() -> None:
    _, builder = round_trip_for(AwsIamUser)
    res = builder.resources_of(AwsIamUser)[0]

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_user"
        assert kwargs["UserName"] == res.name

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.delete_resource(client)


def test_aws_iam_instance_profile_deletion() -> None:
    res, _ = round_trip_for(AwsIamInstanceProfile)

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete_instance_profile"
        assert kwargs["InstanceProfileName"] == res.name

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.delete_resource(client)


def test_tagging() -> None:

    res, _ = round_trip_for(AwsIamServerCertificate, "dns_names", "sha1_fingerprint")

    def validate_update_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "tag_server_certificate"
        assert kwargs["Tags"] == [{"Key": "foo", "Value": "bar"}]
        assert kwargs["ServerCertificateName"] == res.name

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "untag_server_certificate"
        assert kwargs["TagKeys"] == ["foo"]
        assert kwargs["ServerCertificateName"] == res.name

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    res.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    res.delete_resource_tag(client, "foo")



from functools import lru_cache
import logging

import botocore.exceptions
import multiprocessing
from resotolib.baseresources import BaseResource
import resotolib.proc
import resotolib.logger
from resotolib.logger import log, setup_logger
from concurrent import futures
from resotolib.args import ArgumentParser
from resotolib.args import Namespace
from resotolib.config import Config, RunningConfig
from resotolib.graph import Graph
from resotolib.utils import log_runtime
from resotolib.baseplugin import BaseCollectorPlugin
from .config import AwsConfig
from .utils import aws_session
from .resources import AWSAccount, AWSResource
from .accountcollector import AWSAccountCollector
from prometheus_client import Summary, Counter
from typing import List, Optional, cast
from resoto_plugin_aws.aws_client import AwsClient


logging.getLogger("boto").setLevel(logging.CRITICAL)

metrics_collect = Summary("resoto_plugin_aws_collect_seconds", "Time it took the collect() method")
metrics_unhandled_account_exceptions = Counter(
    "resoto_plugin_aws_unhandled_account_exceptions_total",
    "Unhandled AWS Plugin Account Exceptions",
    ["account"],
)


class AWSCollectorPlugin(BaseCollectorPlugin):
    cloud = "aws"

    def __init__(self) -> None:
        super().__init__()
        self.__regions: List[str] = []

    @staticmethod
    def add_config(cfg: Config) -> None:
        cfg.add_config(AwsConfig)

    @metrics_collect.time()  # type: ignore
    def collect(self) -> None:
        log.debug("plugin: AWS collecting resources")

        accounts = get_accounts()
        if len(accounts) == 0:
            log.error("No accounts found")
            return
        for account in accounts:
            add_str = ""
            if account.role:
                add_str += f" role {account.role}"
            if account.profile:
                add_str += f" profile {account.profile}"
            log.debug(f"Found {account.rtdname}{add_str}")

        max_workers = len(accounts) if len(accounts) < Config.aws.account_pool_size else Config.aws.account_pool_size
        pool_args = {"max_workers": max_workers}
        if Config.aws.fork_process:
            pool_args["mp_context"] = multiprocessing.get_context("spawn")
            pool_args["initializer"] = resotolib.proc.initializer
            pool_executor = futures.ProcessPoolExecutor
        else:
            pool_executor = futures.ThreadPoolExecutor  # type: ignore

        with pool_executor(**pool_args) as executor:
            wait_for = [
                executor.submit(
                    collect_account,
                    account,
                    self.regions(profile=account.profile),
                    ArgumentParser.args,
                    Config.running_config,
                )
                for account in accounts
            ]
            for future in futures.as_completed(wait_for):
                account_graph = future.result()
                if not isinstance(account_graph, Graph):
                    log.debug(f"Skipping account graph of invalid type {type(account_graph)}")
                    continue
                self.graph.merge(account_graph)

    def regions(self, profile: Optional[str] = None) -> List[str]:
        if len(self.__regions) == 0:
            if not Config.aws.region or (isinstance(Config.aws.region, list) and len(Config.aws.region) == 0):
                log.debug("AWS region not specified, assuming all regions")
                self.__regions = all_regions(profile=profile)
            else:
                self.__regions = list(Config.aws.region)
        return self.__regions

    @staticmethod
    def update_tag(config: Config, resource: BaseResource, key: str, value: str) -> bool:
        """Update the tag of a resource"""
        if isinstance(resource, AWSResource):  # backwards compatibility with the existing collector
            return resource.update_tag(key, value)

        # TODO: uncomment me when new collector is used
        # if isinstance(resource, AwsResource):
        #     cache_key = int(time.time() / 600)  # 10 minutes
        #     client = get_client(cache_key, config, resource, "Tag update is not possible")
        #     return resource.update_resource_tag(client, key, value)

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")

    @staticmethod
    def delete_tag(config: Config, resource: BaseResource, key: str) -> bool:
        """Delete the tag of a resource"""
        if isinstance(resource, AWSResource):  # backwards compatibility with the existing collector
            return resource.delete_tag(key)

        # TODO: uncomment me when new collector is used
        # if isinstance(resource, AwsResource):
        #     cache_key = int(time.time() / 600)  # 10 minutes
        #     client = get_client(cache_key, config, resource, "Tag deletion is not possible")
        #     return resource.delete_resource_tag(client, key)

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")

    @staticmethod
    def pre_cleanup(config: Config, resource: BaseResource, graph: Graph) -> bool:
        if isinstance(resource, AWSResource):  # backwards compatibility with the existing collector
            return cast(bool, resource.pre_cleanup(graph))  # mypy thinks this isn't a bool

        # TODO: uncomment me when new collector is used
        # if isinstance(resource, AwsResource):
        #     cache_key = int(time.time() / 600)  # 10 minutes
        #     client = get_client(cache_key, config, resource, "Pre-cleanup is not possible")
        #
        #     if not hasattr(resource, "pre_delete_resource"):
        #         return True
        #
        #     if graph is None:
        #         graph = resource._graph
        #
        #     if resource.protected:
        #         log.error(f"Resource {resource.rtdname} is protected - refusing modification")
        #         resource.log(("Modification was requested even though resource is protected" " - refusing"))
        #         return False
        #
        #     if resource.phantom:
        #         raise RuntimeError(f"Can't cleanup phantom resource {resource.rtdname}")
        #
        #     if resource.cleaned:
        #         log.debug(f"Resource {resource.rtdname} has already been cleaned up")
        #         return True
        #
        #     account = resource.account(graph)
        #     region = resource.region(graph)
        #     if not isinstance(account, BaseAccount) or not isinstance(region, BaseRegion):
        #         log.error(("Could not determine account or region for pre cleanup of" f" {resource.rtdname}"))
        #         return False
        #
        #     log_suffix = f" in account {account.dname} region {region.name}"
        #     resource.log("Trying to run pre clean up")
        #     log.debug(f"Trying to run pre clean up {resource.rtdname}{log_suffix}")
        #     try:
        #         if not getattr(resource, "pre_delete_resource")(client, graph):
        #             resource.log("Failed to run pre clean up")
        #             log.error(f"Failed to run pre clean up {resource.rtdname}{log_suffix}")
        #             return False
        #         resource.log("Successfully ran pre clean up")
        #         log.info(f"Successfully ran pre clean up {resource.rtdname}{log_suffix}")
        #     except Exception as e:
        #         resource.log("An error occurred during pre clean up", exception=e)
        #         log.exception(f"An error occurred during pre clean up {resource.rtdname}{log_suffix}")
        #         cloud = resource.cloud(graph)
        #         metrics_resource_pre_cleanup_exceptions.labels(
        #             cloud=cloud.name,
        #             account=account.dname,
        #             region=region.name,
        #             kind=resource.kind,
        #         ).inc()
        #         return False
        #     return True

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")

    @staticmethod
    def cleanup(config: Config, resource: BaseResource, graph: Graph) -> bool:
        if isinstance(resource, AWSResource):  # backwards compatibility with the existing collector
            return cast(bool, resource.cleanup(graph))  # mypy thinks this isn't a bool

        # TODO: uncomment me when new collector is used
        # if isinstance(resource, AwsResource):
        #
        #     cache_key = int(time.time() / 600)  # 10 minutes
        #     client = get_client(cache_key, config, resource, "Pre-cleanup is not possible")
        #
        #     if resource.phantom:
        #         raise RuntimeError(f"Can't cleanup phantom resource {resource.rtdname}")
        #
        #     if resource.cleaned:
        #         log.debug(f"Resource {resource.rtdname} has already been cleaned up")
        #         return True
        #
        #     if resource.protected:
        #         log.error(f"Resource {resource.rtdname} is protected - refusing modification")
        #         resource.log(("Modification was requested even though resource is protected" " - refusing"))
        #         return False
        #
        #     resource._changes.add("cleaned")
        #     if graph is None:
        #         graph = resource._graph
        #
        #     account = resource.account(graph)
        #     region = resource.region(graph)
        #     if not isinstance(account, BaseAccount) or not isinstance(region, BaseRegion):
        #         log.error(f"Could not determine account or region for cleanup of {resource.rtdname}")
        #         return False
        #
        #     log_suffix = f" in account {account.dname} region {region.name}"
        #     resource.log("Trying to clean up")
        #     log.debug(f"Trying to clean up {resource.rtdname}{log_suffix}")
        #     try:
        #         if not resource.delete_resource(client):
        #             resource.log("Failed to clean up")
        #             log.error(f"Failed to clean up {resource.rtdname}{log_suffix}")
        #             return False
        #         resource._cleaned = True
        #         resource.log("Successfully cleaned up")
        #         log.info(f"Successfully cleaned up {resource.rtdname}{log_suffix}")
        #     except Exception as e:
        #         resource.log("An error occurred during clean up", exception=e)
        #         log.exception(f"An error occurred during clean up {resource.rtdname}{log_suffix}")
        #         cloud = resource.cloud(graph)
        #         metrics_resource_cleanup_exceptions.labels(
        #             cloud=cloud.name,
        #             account=account.dname,
        #             region=region.name,
        #             kind=resource.kind,
        #         ).inc()
        #         return False
        #     return True

        raise RuntimeError(f"Unsupported resource type: {resource.rtdname}")


def authenticated(account: AWSAccount) -> bool:
    try:
        log.debug(f"AWS testing credentials for {account.rtdname}")
        session = aws_session(account.id, account.role, account.profile)
        _ = session.client("sts").get_caller_identity().get("Account")
    except botocore.exceptions.NoCredentialsError:
        log.error(f"No AWS credentials found for {account.rtdname}")
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "AuthFailure":
            log.error(f"AWS was unable to validate the provided access credentials for {account.rtdname}")
        elif e.response["Error"]["Code"] == "InvalidClientTokenId":
            log.error(f"AWS was unable to validate the provided security token for {account.rtdname}")
        elif e.response["Error"]["Code"] == "ExpiredToken":
            log.error(f"AWS security token included in the request is expired for {account.rtdname}")
        elif e.response["Error"]["Code"] == "AccessDenied":
            log.error(f"AWS denied access to {account.rtdname}")
        else:
            raise
        return False
    return True


def get_client(cache_key: int, config: Config, resource: BaseResource, err_reason: str) -> AwsClient:
    all_accounts = {acc.id: acc for acc in cached_accounts(cache_key)}
    account = all_accounts.get(resource.account().id)
    if not account:
        msg = f"Unknown account {resource.account().rtdname} in resource {resource.rtdname}. "
        raise RuntimeError(msg + err_reason)

    return AwsClient(config.aws, account.id, role=account.role, profile=account.profile, region=account.region().name)


def current_account_id(profile: Optional[str] = None) -> str:
    session = aws_session(profile=profile)
    return session.client("sts").get_caller_identity().get("Account")  # type: ignore


@lru_cache(maxsize=1024)
def cached_accounts(cache_key: int) -> List[AWSAccount]:
    return get_accounts()


def get_accounts() -> List[AWSAccount]:
    accounts = []
    profiles = [None]

    if Config.aws.assume_current and not Config.aws.do_not_scrape_current:
        raise ValueError(
            "You specified assume_current but not do_not_scrape_current! "
            "This will result in the same account being collected twice and is likely not what you want."
        )

    if isinstance(Config.aws.profiles, list) and len(Config.aws.profiles) > 0:
        log.debug("Using specified AWS profiles")
        profiles = Config.aws.profiles
        if Config.aws.account and len(Config.aws.profiles) > 1:
            raise ValueError(
                "You specified both a list of accounts and more than one profile! "
                "This will result in the attempt to collect the same accounts for "
                "every profile and is likely not what you want."
            )

    for profile in profiles:
        if profile is not None:
            log.debug(f"Finding accounts for profile {profile}")

        try:
            if Config.aws.role and Config.aws.scrape_org:
                log.debug("Role and scrape_org are both set")
                accounts.extend(
                    [
                        AWSAccount(id=aws_account_id, role=Config.aws.role, profile=profile)
                        for aws_account_id in get_org_accounts(
                            filter_current_account=not Config.aws.assume_current, profile=profile
                        )
                        if aws_account_id not in Config.aws.scrape_exclude_account
                    ]
                )
                if not Config.aws.do_not_scrape_current:
                    accounts.append(AWSAccount(id=current_account_id(profile=profile)))
            elif Config.aws.role and Config.aws.account:
                log.debug("Both, role and list of accounts specified")
                accounts.extend(
                    [
                        AWSAccount(id=aws_account_id, role=Config.aws.role, profile=profile)
                        for aws_account_id in Config.aws.account
                    ]
                )
            else:
                accounts.extend([AWSAccount(id=current_account_id(profile=profile), profile=profile)])
        except botocore.exceptions.NoCredentialsError:
            log.error(f"No AWS credentials found for {profile}")
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "AuthFailure":
                log.error(f"AWS was unable to validate the provided access credentials for {profile}")
            elif e.response["Error"]["Code"] == "InvalidClientTokenId":
                log.error(f"AWS was unable to validate the provided security token for {profile}")
            elif e.response["Error"]["Code"] == "ExpiredToken":
                log.error(f"AWS security token included in the request is expired for {profile}")
            elif e.response["Error"]["Code"] == "AccessDenied":
                log.error(f"AWS denied access for {profile}")
            else:
                raise

    return accounts


def get_org_accounts(filter_current_account: bool = False, profile: Optional[str] = None) -> List[str]:
    session = aws_session(profile=profile)
    client = session.client("organizations")
    accounts = []
    try:
        response = client.list_accounts()
        accounts = response.get("Accounts", [])
        while response.get("NextToken") is not None:
            response = client.list_accounts(NextToken=response["NextToken"])
            accounts.extend(response.get("Accounts", []))
    except botocore.exceptions.ClientError as e:
        if e.response["Error"]["Code"] == "AccessDeniedException":
            log.error("AWS error - missing permissions to list organization accounts")
        else:
            raise
    filter_account_id = current_account_id(profile=profile) if filter_current_account else -1
    accounts = [aws_account["Id"] for aws_account in accounts if aws_account["Id"] != filter_account_id]
    for account in accounts:
        log.debug(f"AWS found org account {account}")
    log.info(f"AWS found a total of {len(accounts)} org accounts")
    return accounts


def all_regions(profile: Optional[str] = None) -> List[str]:
    session = aws_session(profile=profile)
    ec2 = session.client("ec2", region_name="us-east-1")
    regions = ec2.describe_regions()
    return [r["RegionName"] for r in regions["Regions"]]


@log_runtime  # type: ignore
def collect_account(
    account: AWSAccount,
    regions: List[str],
    args: Namespace,
    running_config: RunningConfig,
) -> Optional[Graph]:
    collector_name = f"aws_{account.id}"
    resotolib.proc.set_thread_name(collector_name)

    if args is not None:
        ArgumentParser.args = args
        setup_logger("resotoworker-aws")
    if running_config is not None:
        Config.running_config.apply(running_config)

    if not authenticated(account):
        log.error(f"Skipping {account.rtdname} due to authentication failure")
        return None

    log.debug(f"Starting new collect process for account {account.dname}")

    aac = AWSAccountCollector(regions, account)
    try:
        aac.collect()
    except botocore.exceptions.ClientError as e:
        log.exception(f"An AWS {e.response['Error']['Code']} error occurred while collecting account {account.dname}")
        metrics_unhandled_account_exceptions.labels(account=account.dname).inc()
        return None
    except Exception:
        log.exception(f"An unhandled error occurred while collecting AWS account {account.dname}")
        metrics_unhandled_account_exceptions.labels(account=account.dname).inc()
        return None

    return aac.graph  # type: ignore



import sys
import resotolib.logger
from resotolib.args import get_arg_parser, ArgumentParser
from resoto_plugin_aws import get_org_accounts


resotolib.logger.getLogger("resoto").setLevel(resotolib.logger.ERROR)
log = resotolib.logger.getLogger(__name__)

argv = sys.argv[1:]
if "-v" in argv or "--verbose" in argv:
    resotolib.logger.getLogger("resoto").setLevel(resotolib.logger.DEBUG)


def main() -> None:
    # Add cli args
    arg_parser = get_arg_parser()
    add_args(arg_parser)
    arg_parser.parse_args()
    accounts = get_org_accounts()
    for account in accounts:
        print(account)


def add_args(arg_parser: ArgumentParser) -> None:
    arg_parser.add_argument("--aws-access-key-id", help="AWS Access Key ID", dest="aws_access_key_id")
    arg_parser.add_argument(
        "--aws-secret-access-key",
        help="AWS Secret Access Key",
        dest="aws_secret_access_key",
    )
    arg_parser.add_argument("--aws-role", help="AWS IAM Role", dest="aws_role")
    arg_parser.add_argument(
        "--aws-role-override",
        help="Override any stored roles (e.g. from remote graphs) (default: False)",
        dest="aws_role_override",
        action="store_true",
        default=False,
    )


if __name__ == "__main__":
    main()



import re
from typing import ClassVar, Dict, Optional, List, Type

from attrs import define, field
from resoto_plugin_aws.aws_client import AwsClient

from resoto_plugin_aws.resource.base import AwsResource, GraphBuilder, AwsApiSpec
from resoto_plugin_aws.resource.ec2 import AwsEc2Subnet, AwsEc2SecurityGroup, AwsEc2Vpc
from resotolib.baseresources import (  # noqa: F401
    BaseCertificate,
    BasePolicy,
    BaseGroup,
    BaseAccount,
    BaseAccessKey,
    BaseUser,
    BaseServerlessFunction,
    ModelReference,
)
from resotolib.json_bender import Bender, S, Bend, ForallBend, F
from resotolib.types import Json


@define(eq=False, slots=False)
class AwsLambdaEnvironmentError:
    kind: ClassVar[str] = "aws_lambda_environment_error"
    mapping: ClassVar[Dict[str, Bender]] = {"error_code": S("ErrorCode"), "message": S("Message")}
    error_code: Optional[str] = field(default=None)
    message: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsLambdaEnvironmentResponse:
    kind: ClassVar[str] = "aws_lambda_environment_response"
    mapping: ClassVar[Dict[str, Bender]] = {
        "variables": S("Variables"),
        "error": S("Error") >> Bend(AwsLambdaEnvironmentError.mapping),
    }
    variables: Optional[Dict[str, str]] = field(default=None)
    error: Optional[AwsLambdaEnvironmentError] = field(default=None)


@define(eq=False, slots=False)
class AwsLambdaLayer:
    kind: ClassVar[str] = "aws_lambda_layer"
    mapping: ClassVar[Dict[str, Bender]] = {
        "arn": S("Arn"),
        "code_size": S("CodeSize"),
        "signing_profile_version_arn": S("SigningProfileVersionArn"),
        "signing_job_arn": S("SigningJobArn"),
    }
    arn: Optional[str] = field(default=None)
    code_size: Optional[int] = field(default=None)
    signing_profile_version_arn: Optional[str] = field(default=None)
    signing_job_arn: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsLambdaFileSystemConfig:
    kind: ClassVar[str] = "aws_lambda_file_system_config"
    mapping: ClassVar[Dict[str, Bender]] = {"arn": S("Arn"), "local_mount_path": S("LocalMountPath")}
    arn: Optional[str] = field(default=None)
    local_mount_path: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsLambdaImageConfig:
    kind: ClassVar[str] = "aws_lambda_image_config"
    mapping: ClassVar[Dict[str, Bender]] = {
        "entry_point": S("EntryPoint", default=[]),
        "command": S("Command", default=[]),
        "working_directory": S("WorkingDirectory"),
    }
    entry_point: List[str] = field(factory=list)
    command: List[str] = field(factory=list)
    working_directory: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsLambdaImageConfigError:
    kind: ClassVar[str] = "aws_lambda_image_config_error"
    mapping: ClassVar[Dict[str, Bender]] = {"error_code": S("ErrorCode"), "message": S("Message")}
    error_code: Optional[str] = field(default=None)
    message: Optional[str] = field(default=None)


@define(eq=False, slots=False)
class AwsLambdaImageConfigResponse:
    kind: ClassVar[str] = "aws_lambda_image_config_response"
    mapping: ClassVar[Dict[str, Bender]] = {
        "image_config": S("ImageConfig") >> Bend(AwsLambdaImageConfig.mapping),
        "error": S("Error") >> Bend(AwsLambdaImageConfigError.mapping),
    }
    image_config: Optional[AwsLambdaImageConfig] = field(default=None)
    error: Optional[AwsLambdaImageConfigError] = field(default=None)


@define(eq=False, slots=False)
class AwsLambdaFunction(AwsResource, BaseServerlessFunction):
    kind: ClassVar[str] = "aws_lambda_function"
    api_spec: ClassVar[AwsApiSpec] = AwsApiSpec("lambda", "list-functions", "Functions")
    reference_kinds: ClassVar[ModelReference] = {
        "predecessors": {
            "default": ["aws_vpc", "aws_ec2_subnet", "aws_ec2_security_group"],
            "delete": ["aws_vpc", "aws_ec2_subnet"],
        },
    }
    mapping: ClassVar[Dict[str, Bender]] = {
        "id": S("FunctionName"),
        "name": S("FunctionName"),
        "mtime": S("LastModified") >> F(lambda x: re.sub(r"(\d\d)(\d\d)$", "\\1:\\2", x)),  # time zone is without colon
        "arn": S("FunctionArn"),
        "function_runtime": S("Runtime"),
        "function_role": S("Role"),
        "function_handler": S("Handler"),
        "function_code_size": S("CodeSize"),
        "function_description": S("Description"),
        "function_timeout": S("Timeout"),
        "function_memory_size": S("MemorySize"),
        "function_code_sha256": S("CodeSha256"),
        "function_version": S("Version"),
        "function_dead_letter_config": S("DeadLetterConfig", "TargetArn"),
        "function_environment": S("Environment") >> Bend(AwsLambdaEnvironmentResponse.mapping),
        "function_kms_key_arn": S("KMSKeyArn"),
        "function_tracing_config": S("TracingConfig", "Mode"),
        "function_master_arn": S("MasterArn"),
        "function_revision_id": S("RevisionId"),
        "function_layers": S("Layers", default=[]) >> ForallBend(AwsLambdaLayer.mapping),
        "function_state": S("State"),
        "function_state_reason": S("StateReason"),
        "function_state_reason_code": S("StateReasonCode"),
        "function_last_update_status": S("LastUpdateStatus"),
        "function_last_update_status_reason": S("LastUpdateStatusReason"),
        "function_last_update_status_reason_code": S("LastUpdateStatusReasonCode"),
        "function_file_system_configs": S("FileSystemConfigs", default=[])
        >> ForallBend(AwsLambdaFileSystemConfig.mapping),
        "function_package_type": S("PackageType"),
        "function_image_config_response": S("ImageConfigResponse") >> Bend(AwsLambdaImageConfigResponse.mapping),
        "function_signing_profile_version_arn": S("SigningProfileVersionArn"),
        "function_signing_job_arn": S("SigningJobArn"),
        "function_architectures": S("Architectures", default=[]),
        "function_ephemeral_storage": S("EphemeralStorage", "Size"),
    }
    function_runtime: Optional[str] = field(default=None)
    function_role: Optional[str] = field(default=None)
    function_handler: Optional[str] = field(default=None)
    function_code_size: Optional[int] = field(default=None)
    function_description: Optional[str] = field(default=None)
    function_timeout: Optional[int] = field(default=None)
    function_memory_size: Optional[int] = field(default=None)
    function_code_sha256: Optional[str] = field(default=None)
    function_version: Optional[str] = field(default=None)
    function_dead_letter_config: Optional[str] = field(default=None)
    function_environment: Optional[AwsLambdaEnvironmentResponse] = field(default=None)
    function_kms_key_arn: Optional[str] = field(default=None)
    function_tracing_config: Optional[str] = field(default=None)
    function_master_arn: Optional[str] = field(default=None)
    function_revision_id: Optional[str] = field(default=None)
    function_layers: List[AwsLambdaLayer] = field(factory=list)
    function_state: Optional[str] = field(default=None)
    function_state_reason: Optional[str] = field(default=None)
    function_state_reason_code: Optional[str] = field(default=None)
    function_last_update_status: Optional[str] = field(default=None)
    function_last_update_status_reason: Optional[str] = field(default=None)
    function_last_update_status_reason_code: Optional[str] = field(default=None)
    function_file_system_configs: List[AwsLambdaFileSystemConfig] = field(factory=list)
    function_package_type: Optional[str] = field(default=None)
    function_image_config_response: Optional[AwsLambdaImageConfigResponse] = field(default=None)
    function_signing_profile_version_arn: Optional[str] = field(default=None)
    function_signing_job_arn: Optional[str] = field(default=None)
    function_architectures: List[str] = field(factory=list)
    function_ephemeral_storage: Optional[int] = field(default=None)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        if vpc_config := source.get("VpcConfig"):
            if vpc_id := vpc_config.get("VpcId"):
                builder.dependant_node(self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Vpc, id=vpc_id)
            for subnet_id in vpc_config.get("SubnetIds", []):
                builder.dependant_node(
                    self, reverse=True, delete_same_as_default=True, clazz=AwsEc2Subnet, id=subnet_id
                )
            for security_group_id in vpc_config.get("SecurityGroupIds", []):
                builder.add_edge(self, reverse=True, clazz=AwsEc2SecurityGroup, id=security_group_id)

    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        client.call(
            service=self.api_spec.service,
            action="tag_resource",
            result_name=None,
            Resource=self.arn,
            Tags={key: value},
        )
        return True

    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        client.call(
            service=self.api_spec.service,
            action="untag_resource",
            result_name=None,
            Resource=self.arn,
            TagKeys=[key],
        )
        return True

    def delete_resource(self, client: AwsClient) -> bool:
        client.call(service=self.api_spec.service, action="delete_function", result_name=None, FunctionName=self.arn)
        return True


resources: List[Type[AwsResource]] = [AwsLambdaFunction]



from types import SimpleNamespace
from typing import Any, cast
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.sqs import AwsSqsQueue
from test.resources import round_trip_for


def test_queues() -> None:
    first, builder = round_trip_for(AwsSqsQueue)
    assert len(builder.resources_of(AwsSqsQueue)) == 1
    assert len(first.tags) == 2


def test_tagging_queues() -> None:
    queue, _ = round_trip_for(AwsSqsQueue)

    def validate_update_args(**kwargs: Any) -> Any:
        if kwargs["action"] == "tag-queue":
            assert kwargs["QueueUrl"] == queue.sqs_queue_url
            assert kwargs["Tags"] == {"foo": "bar"}

    def validate_delete_args(**kwargs: Any) -> Any:
        if kwargs["action"] == "untag-queue":
            assert kwargs["QueueUrl"] == queue.sqs_queue_url
            assert kwargs["TagKeys"] == ["foo"]

    client = cast(AwsClient, SimpleNamespace(call=validate_update_args))
    queue.update_resource_tag(client, "foo", "bar")

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    queue.delete_resource_tag(client, "foo")


def test_delete_queues() -> None:
    queue, _ = round_trip_for(AwsSqsQueue)

    def validate_delete_args(**kwargs: Any) -> None:
        assert kwargs["action"] == "delete-queue"
        assert kwargs["QueueUrl"] == queue.sqs_queue_url

    client = cast(AwsClient, SimpleNamespace(call=validate_delete_args))
    queue.delete_resource(client)



from __future__ import annotations

import concurrent
import logging
from concurrent.futures import Executor, Future
from datetime import datetime, timezone
from functools import lru_cache
from threading import Lock
from typing import ClassVar, Dict, Optional, List, Type, Any, TypeVar, Callable

from attr import evolve
from attrs import define
from boto3.exceptions import Boto3Error

from resoto_plugin_aws.config import AwsConfig
from resoto_plugin_aws.aws_client import AwsClient
from resoto_plugin_aws.resource.pricing import AwsPricingPrice
from resotolib.baseresources import (
    BaseResource,
    EdgeType,
    Cloud,
    BaseAccount,
    BaseRegion,
    BaseVolumeType,
    ModelReference,
)
from resotolib.graph import Graph
from resotolib.json import to_json as to_js, from_json as from_js
from resotolib.json_bender import Bender, bend
from resotolib.types import Json

log = logging.getLogger("resoto.plugins.aws")


@define
class AwsApiSpec:
    """
    Specifications for the AWS API to call and the expected response.
    """

    service: str
    api_action: str
    result_property: Optional[str] = None
    parameter: Optional[Dict[str, Any]] = None


@define(eq=False, slots=False)
class AwsResource(BaseResource):
    """
    Base class for all AWS resources.
    Override kind, mapping and api_spec for every resource that is collected in AWS.
    """

    # The name of the kind of all resources. Needs to be globally unique.
    kind: ClassVar[str] = "aws_resource"
    # The mapping to transform the incoming API json into the internal representation.
    mapping: ClassVar[Dict[str, Bender]] = {}
    # Which API to call and what to expect in the result.
    api_spec: ClassVar[Optional[AwsApiSpec]] = None

    # The AWS specific identifier of the resource. Not available for all resources.
    arn: Optional[str] = None

    # TODO: implement me
    def update_resource_tag(self, client: AwsClient, key: str, value: str) -> bool:
        raise NotImplementedError

    # TODO: implement me
    def delete_resource_tag(self, client: AwsClient, key: str) -> bool:
        raise NotImplementedError

    # legacy interface
    def delete(self, graph: Graph) -> bool:
        raise NotImplementedError

    def delete_resource(self, client: AwsClient) -> bool:
        return False

    def to_json(self) -> Json:
        return to_js(
            self,
            strip_attr=(
                "mapping",
                "phantom",
                "reference_kinds",
                "parent_resource",
                "usage_percentage",
                "dname",
                "kdname",
                "rtdname",
                "changes",
                "event_log",
                "str_event_log",
                "chksum",
                "age",
                "last_access",
                "last_update",
                "clean",
                "cleaned",
                "protected",
                "_graph",
                "graph",
                "max_graph_depth",
                "resource_type",
                "age",
                "last_access",
                "last_update",
                "clean",
                "cleaned",
                "protected",
                "uuid",
                "kind",
                "api_spec",
            ),
        )

    @classmethod
    def from_json(cls: Type[AwsResourceType], json: Json) -> AwsResourceType:
        return from_js(json, cls)

    @classmethod
    def from_api(cls: Type[AwsResourceType], json: Json) -> AwsResourceType:
        mapped = bend(cls.mapping, json)
        return cls.from_json(mapped)

    @classmethod
    def collect_resources(cls: Type[AwsResource], builder: GraphBuilder) -> None:
        # Default behavior: in case the class has an ApiSpec, call the api and call collect.
        log.debug(f"Collecting {cls.__name__} in region {builder.region.name}")
        if spec := cls.api_spec:
            try:
                kwargs = spec.parameter or {}
                items = builder.client.list(spec.service, spec.api_action, spec.result_property, **kwargs)
                cls.collect(items, builder)
            except Boto3Error as e:
                log.error(f"Error while collecting {cls.__name__} in region {builder.region.name}: {e}")
                raise

    @classmethod
    def collect(cls: Type[AwsResource], json: List[Json], builder: GraphBuilder) -> None:
        # Default behavior: iterate over json snippets and for each:
        # - bend the json
        # - transform the result into a resource
        # - add the resource to the graph
        # In case additional work needs to be done, override this method.
        for js in json:
            instance = cls.from_api(js)
            builder.add_node(instance, js)

    def connect_in_graph(self, builder: GraphBuilder, source: Json) -> None:
        # Default behavior: add resource to the namespace
        pass

    def __str__(self) -> str:
        return f"{self.kind}:{self.name}"


AwsResourceType = TypeVar("AwsResourceType", bound=AwsResource)


# derived from https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-template-resource-type-ref.html
@define(eq=False)
class AwsAccount(BaseAccount, AwsResource):
    kind: ClassVar[str] = "aws_account"
    reference_kinds: ClassVar[ModelReference] = {"successors": {"default": ["aws_region"]}}

    account_alias: Optional[str] = ""
    role: Optional[str] = None
    profile: Optional[str] = None
    users: Optional[int] = 0
    groups: Optional[int] = 0
    account_mfa_enabled: Optional[int] = 0
    account_access_keys_present: Optional[int] = 0
    account_signing_certificates_present: Optional[int] = 0
    mfa_devices: Optional[int] = 0
    mfa_devices_in_use: Optional[int] = 0
    policies: Optional[int] = 0
    policy_versions_in_use: Optional[int] = 0
    global_endpoint_token_version: Optional[int] = 0
    server_certificates: Optional[int] = 0
    minimum_password_length: Optional[int] = None
    require_symbols: Optional[bool] = None
    require_numbers: Optional[bool] = None
    require_uppercase_characters: Optional[bool] = None
    require_lowercase_characters: Optional[bool] = None
    allow_users_to_change_password: Optional[bool] = None
    expire_passwords: Optional[bool] = None
    max_password_age: Optional[int] = 0
    password_reuse_prevention: Optional[int] = 0
    hard_expiry: Optional[bool] = None


default_ctime = datetime(2006, 3, 19, tzinfo=timezone.utc)  # AWS public launch date


@define(eq=False)
class AwsRegion(BaseRegion, AwsResource):
    kind: ClassVar[str] = "aws_region"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [
                "aws_vpc_quota",
                "aws_vpc_peering_connection",
                "aws_vpc_endpoint",
                "aws_vpc",
                "aws_s3_bucket_quota",
                "aws_s3_bucket",
                "aws_rds_instance",
                "aws_iam_server_certificate_quota",
                "aws_iam_server_certificate",
                "aws_iam_role",
                "aws_iam_policy",
                "aws_iam_instance_profile",
                "aws_iam_group",
                "aws_elb_quota",
                "aws_elb",
                "aws_eks_cluster",
                "aws_ec2_volume_type",
                "aws_ec2_volume",
                "aws_iam_user",
                "aws_ec2_subnet",
                "aws_ec2_snapshot",
                "aws_ec2_security_group",
                "aws_ec2_route_table",
                "aws_ec2_network_interface",
                "aws_ec2_network_acl",
                "aws_ec2_nat_gateway",
                "aws_ec2_keypair",
                "aws_ec2_internet_gateway_quota",
                "aws_ec2_internet_gateway",
                "aws_ec2_instance_type",
                "aws_ec2_instance_quota",
                "aws_ec2_instance",
                "aws_ec2_elastic_ip",
                "aws_cloudwatch_alarm",
                "aws_cloudformation_stack",
                "aws_cloudformation_stack_set",
                "aws_autoscaling_group",
                "aws_alb_target_group",
                "aws_alb_quota",
                "aws_alb",
            ]
        }
    }
    ctime: Optional[datetime] = default_ctime


@define(eq=False, slots=False)
class AwsEc2VolumeType(AwsResource, BaseVolumeType):
    kind: ClassVar[str] = "aws_ec2_volume_type"


@define
class ExecutorQueue:
    executor: Executor
    name: str
    futures: List[Future[Any]] = []
    _lock: Lock = Lock()

    def submit_work(self, fn: Callable[..., None], *args: Any, **kwargs: Any) -> Future[Any]:
        future = self.executor.submit(fn, *args, **kwargs)
        with self._lock:
            self.futures.append(future)
        return future

    def wait_for_submitted_work(self) -> None:
        # wait until all futures are complete
        with self._lock:
            to_wait = self.futures
            self.futures = []
        for future in concurrent.futures.as_completed(to_wait):
            try:
                future.result()
            except Exception as ex:
                log.exception(f"Unhandled exception in account {self.name}: {ex}")
                raise


class GraphBuilder:
    def __init__(
        self,
        graph: Graph,
        cloud: Cloud,
        account: AwsAccount,
        region: AwsRegion,
        client: AwsClient,
        executor: ExecutorQueue,
        global_instance_types: Optional[Dict[str, Any]] = None,
    ) -> None:
        self.graph = graph
        self.cloud = cloud
        self.account = account
        self.region = region
        self.client = client
        self.executor = executor
        self.name = f"AWS:{account.name}:{region.name}"
        self.global_instance_types: Dict[str, Any] = global_instance_types or {}

    def submit_work(self, fn: Callable[..., None], *args: Any, **kwargs: Any) -> Future[Any]:
        return self.executor.submit_work(fn, *args, **kwargs)

    @property
    def config(self) -> AwsConfig:
        return self.client.config

    def node(self, clazz: Optional[Type[AwsResourceType]] = None, **node: Any) -> Optional[AwsResourceType]:
        if isinstance(nd := node.get("node"), AwsResource):
            return nd  # type: ignore
        for n in self.graph:
            is_clazz = isinstance(n, clazz) if clazz else True
            if is_clazz and all(getattr(n, k, None) == v for k, v in node.items()):
                return n  # type: ignore
        return None

    def add_node(self, node: AwsResourceType, source: Optional[Json] = None) -> AwsResourceType:
        log.debug(f"{self.name}: add node {node}")
        node._cloud = self.cloud
        node._account = self.account
        node._region = self.region
        self.graph.add_node(node, source=source or {})
        return node

    def add_edge(
        self, from_node: BaseResource, edge_type: EdgeType = EdgeType.default, reverse: bool = False, **to_node: Any
    ) -> None:
        to_n = self.node(**to_node)
        if isinstance(from_node, AwsResource) and isinstance(to_n, AwsResource):
            start, end = (to_n, from_node) if reverse else (from_node, to_n)
            log.debug(f"{self.name}: add edge: {start} -> {end} [{edge_type}]")
            self.graph.add_edge(start, end, edge_type=edge_type)

    def dependant_node(
        self, from_node: BaseResource, reverse: bool = False, delete_same_as_default: bool = False, **to_node: Any
    ) -> None:
        to_n = self.node(**to_node)
        if isinstance(from_node, AwsResource) and isinstance(to_n, AwsResource):
            start, end = (to_n, from_node) if reverse else (from_node, to_n)
            log.debug(f"{self.name}: add edge: {start} -> {end} [default]")
            self.graph.add_edge(start, end, edge_type=EdgeType.default)
            if delete_same_as_default:
                start, end = end, start
            log.debug(f"{self.name}: add edge: {end} -> {start} [delete]")
            self.graph.add_edge(end, start, edge_type=EdgeType.delete)

    def resources_of(self, resource_type: Type[AwsResourceType]) -> List[AwsResourceType]:
        return [n for n in self.graph.nodes if isinstance(n, resource_type)]

    @lru_cache(maxsize=None)
    def instance_type(self, instance_type: str) -> Optional[Any]:
        if (global_type := self.global_instance_types.get(instance_type)) is None:
            return None  # instance type not found
        price = AwsPricingPrice.instance_type_price(self.client, instance_type, self.region.name)
        return evolve(global_type, region=self.region, ondemand_cost=price.on_demand_price_usd if price else None)

    @lru_cache(maxsize=None)
    def volume_type(self, volume_type: str) -> Optional[Any]:
        price = AwsPricingPrice.volume_type_price(self.client, volume_type, self.region.name)
        return AwsEc2VolumeType(
            id=volume_type,
            name=volume_type,
            volume_type=volume_type,
            ondemand_cost=price.on_demand_price_usd if price else 0,
        )

    def for_region(self, region: AwsRegion) -> GraphBuilder:
        return GraphBuilder(
            self.graph,
            self.cloud,
            self.account,
            region,
            self.client.for_region(region.name),
            self.executor,
            self.global_instance_types,
        )



import time
import copy
import botocore.exceptions
from datetime import date
from enum import auto
from resotolib.baseresources import *
from resotolib.graph import Graph
from resotolib.utils import make_valid_timestamp
from .utils import aws_client, aws_resource, tags_as_dict
from typing import ClassVar, Any, Optional, Union, List, Dict
from attrs import define
from resotolib.logger import log


default_ctime = make_valid_timestamp(date(2006, 3, 19))  # AWS public launch date


# derived from https://docs.aws.amazon.com/AWSCloudFormation/latest/UserGuide/aws-template-resource-type-ref.html
@define(eq=False, slots=False)
class AWSAccount(BaseAccount):
    kind: ClassVar[str] = "aws_account"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_region"],
            "delete": [],
        }
    }

    account_alias: Optional[str] = ""
    role: Optional[str] = None
    profile: Optional[str] = None
    users: Optional[int] = 0
    groups: Optional[int] = 0
    account_mfa_enabled: Optional[int] = 0
    account_access_keys_present: Optional[int] = 0
    account_signing_certificates_present: Optional[int] = 0
    mfa_devices: Optional[int] = 0
    mfa_devices_in_use: Optional[int] = 0
    policies: Optional[int] = 0
    policy_versions_in_use: Optional[int] = 0
    global_endpoint_token_version: Optional[int] = 0
    server_certificates: Optional[int] = 0
    minimum_password_length: Optional[int] = None
    require_symbols: Optional[bool] = None
    require_numbers: Optional[bool] = None
    require_uppercase_characters: Optional[bool] = None
    require_lowercase_characters: Optional[bool] = None
    allow_users_to_change_password: Optional[bool] = None
    expire_passwords: Optional[bool] = None
    max_password_age: Optional[int] = 0
    password_reuse_prevention: Optional[int] = 0
    hard_expiry: Optional[bool] = None

    def delete(self, graph) -> bool:
        return False


@define(eq=False, slots=False)
class AWSRegion(BaseRegion):
    kind: ClassVar[str] = "aws_region"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [
                "aws_vpc_quota",
                "aws_vpc_peering_connection",
                "aws_vpc_endpoint",
                "aws_vpc",
                "aws_s3_bucket_quota",
                "aws_s3_bucket",
                "aws_rds_instance",
                "aws_iam_server_certificate_quota",
                "aws_iam_server_certificate",
                "aws_iam_role",
                "aws_iam_policy",
                "aws_iam_instance_profile",
                "aws_iam_group",
                "aws_elb_quota",
                "aws_elb",
                "aws_eks_cluster",
                "aws_ec2_volume_type",
                "aws_ec2_volume",
                "aws_iam_user",
                "aws_ec2_subnet",
                "aws_ec2_snapshot",
                "aws_ec2_security_group",
                "aws_ec2_route_table",
                "aws_ec2_network_interface",
                "aws_ec2_network_acl",
                "aws_ec2_nat_gateway",
                "aws_ec2_keypair",
                "aws_ec2_internet_gateway_quota",
                "aws_ec2_internet_gateway",
                "aws_ec2_instance_type",
                "aws_ec2_instance_quota",
                "aws_ec2_instance",
                "aws_ec2_elastic_ip",
                "aws_cloudwatch_alarm",
                "aws_cloudformation_stack",
                "aws_cloudformation_stack_set",
                "aws_autoscaling_group",
                "aws_alb_target_group",
                "aws_alb_quota",
                "aws_alb",
                "aws_lambda_function",
            ],
            "delete": [],
        }
    }
    ctime: Optional[datetime] = default_ctime

    def delete(self, graph) -> bool:
        return False


@define(eq=False, slots=False)
class AWSResource:
    kind: ClassVar[str] = "aws_resource"
    arn: Optional[str] = None

    def delete(self, graph) -> bool:
        return False


@define(eq=False, slots=False)
class AWSEC2InstanceType(AWSResource, BaseInstanceType):
    kind: ClassVar[str] = "aws_ec2_instance_type"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_instance"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSEC2InstanceQuota(AWSResource, BaseInstanceQuota):
    kind: ClassVar[str] = "aws_ec2_instance_quota"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_instance_type"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSEC2Instance(AWSResource, BaseInstance):
    kind: ClassVar[str] = "aws_ec2_instance"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [
                "aws_ec2_volume",
                "aws_ec2_network_interface",
                "aws_ec2_keypair",
                "aws_ec2_elastic_ip",
                "aws_cloudwatch_alarm",
            ],
            "delete": [
                "aws_elb",
                "aws_autoscaling_group",
                "aws_alb_target_group",
                "aws_cloudwatch_alarm",
            ],
        }
    }

    def delete(self, graph: Graph) -> bool:
        if self.instance_status == InstanceStatus.TERMINATED:
            log.debug(
                (
                    f"AWS EC2 Instance {self.dname} in"
                    f" account {self.account(graph).dname}"
                    f" region {self.region(graph).name}"
                    " is already terminated"
                )
            )
            self.log("Instance is already terminated")
            return True
        ec2 = aws_resource(self, "ec2", graph)
        instance = ec2.Instance(self.id)
        instance.terminate()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_resource(self, "ec2")
        instance = ec2.Instance(self.id)
        instance.create_tags(Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_resource(self, "ec2")
        instance = ec2.Instance(self.id)
        instance.delete_tags(Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSLambdaFunction(AWSResource, BaseServerlessFunction):
    kind: ClassVar[str] = "aws_lambda_function"
    successor_kinds: ClassVar[Dict[str, List[str]]] = {
        "default": [],
        "delete": [],
    }

    role: Optional[str] = None
    runtime: Optional[str] = None
    code_size: Optional[int] = None
    memory_size: Optional[int] = None
    kms_key_arn: Optional[str] = None

    def delete(self, graph: Graph) -> bool:
        client = aws_client(self, "lambda", graph)
        client.delete_function(FunctionName=self.id)
        return True

    def update_tag(self, key, value) -> bool:
        client = aws_client(self, "lambda")
        client.tag_resource(Resource=self.id, Tags={key: value})
        return True

    def delete_tag(self, key) -> bool:
        client = aws_client(self, "lambda")
        client.untag_resource(Resource=self.id, TagKeys=[key])
        return True


@define(eq=False, slots=False)
class AWSEC2KeyPair(AWSResource, BaseKeyPair):
    kind: ClassVar[str] = "aws_ec2_keypair"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [],
            "delete": ["aws_ec2_instance"],
        }
    }

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_client(self, "ec2", graph)
        ec2.delete_key_pair(KeyName=self.name)
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2VolumeType(AWSResource, BaseVolumeType):
    kind: ClassVar[str] = "aws_ec2_volume_type"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_volume"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSEC2Volume(AWSResource, BaseVolume):
    kind: ClassVar[str] = "aws_ec2_volume"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_snapshot"],
            "delete": ["aws_ec2_instance"],
        }
    }

    volume_kms_key_id: Optional[str] = None
    volume_multi_attach_enabled: Optional[bool] = None
    volume_outpost_arn: Optional[str] = None
    volume_snapshot_id: Optional[str] = None

    def delete(
        self,
        graph: Graph,
        snapshot_before_delete: bool = False,
        snapshot_timeout: int = 3600,
    ) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        volume = ec2.Volume(self.id)
        if snapshot_before_delete or self.snapshot_before_delete:
            log_msg = "Creating snapshot before deletion"
            self.log(log_msg)
            log.debug(f"{log_msg} of {self.kind} {self.dname}")
            snapshot = volume.create_snapshot(
                Description=f"resoto created snapshot for volume {self.id}",
                TagSpecifications=[
                    {
                        "ResourceType": "snapshot",
                        "Tags": [
                            {"Key": "Name", "Value": f"CK snap of {self.id}"},
                            {"Key": "owner", "Value": "resoto"},
                        ],
                    },
                ],
            )
            start_utime = time.time()
            while snapshot.state == "pending":
                if time.time() > start_utime + snapshot_timeout:
                    raise TimeoutError(
                        (
                            f"AWS EC2 Volume Snapshot {self.dname} tag update timed out after "
                            f"{snapshot_timeout} seconds with status {snapshot.state} ({snapshot.state_message})"
                        )
                    )
                time.sleep(10)
                log.debug(
                    (
                        f"Waiting for snapshot {snapshot.id} to finish before deletion of "
                        f"{self.kind} {self.dname} - progress {snapshot.progress}"
                    )
                )
                snapshot = ec2.Snapshot(snapshot.id)
            if snapshot.state != "completed":
                log_msg = f"Failed to create snapshot - status {snapshot.state} ({snapshot.state_message})"
                self.log(log_msg)
                log.error(
                    (
                        f"{log_msg} for {self.kind} {self.dname} in "
                        f"account {self.account(graph).dname} region {self.region(graph).name}"
                    )
                )
                return False
        volume.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_resource(self, "ec2")
        volume = ec2.Volume(self.id)
        volume.create_tags(Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2Snapshot(AWSResource, BaseSnapshot):
    kind: ClassVar[str] = "aws_ec2_snapshot"

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        snapshot = ec2.Snapshot(self.id)
        snapshot.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2Subnet(AWSResource, BaseSubnet):
    kind: ClassVar[str] = "aws_ec2_subnet"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [
                "aws_vpc_endpoint",
                "aws_rds_instance",
                "aws_elb",
                "aws_ec2_network_interface",
                "aws_ec2_network_acl",
                "aws_ec2_nat_gateway",
                "aws_alb",
                "aws_lambda_function",
            ],
            "delete": [
                "aws_vpc_endpoint",
                "aws_rds_instance",
                "aws_elb",
                "aws_ec2_network_interface",
                "aws_alb",
                "aws_lambda_function",
            ],
        }
    }

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        subnet = ec2.Subnet(self.id)
        subnet.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2ElasticIP(AWSResource, BaseIPAddress):
    kind: ClassVar[str] = "aws_ec2_elastic_ip"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [],
            "delete": ["aws_ec2_network_interface", "aws_ec2_instance"],
        }
    }

    instance_id: Optional[str] = None
    public_ip: Optional[str] = None
    allocation_id: Optional[str] = None
    association_id: Optional[str] = None
    domain: Optional[str] = None
    network_interface_id: Optional[str] = None
    network_interface_owner_id: Optional[str] = None
    private_ip_address: Optional[str] = None

    def pre_delete(self, graph: Graph) -> bool:
        if self.association_id is not None:
            ec2 = aws_client(self, "ec2", graph=graph)
            ec2.disassociate_address(AssociationId=self.association_id)
        else:
            log.debug(f"No association for {self.rtdname}")
        return True

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_client(self, "ec2", graph=graph)
        ec2.release_address(AllocationId=self.allocation_id)
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSVPC(AWSResource, BaseNetwork):
    kind: ClassVar[str] = "aws_vpc"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [
                "aws_vpc_peering_connection",
                "aws_vpc_endpoint",
                "aws_rds_instance",
                "aws_elb",
                "aws_ec2_subnet",
                "aws_ec2_security_group",
                "aws_ec2_route_table",
                "aws_ec2_network_interface",
                "aws_ec2_network_acl",
                "aws_ec2_nat_gateway",
                "aws_ec2_internet_gateway",
                "aws_alb_target_group",
                "aws_lambda_function",
            ],
            "delete": [
                "aws_vpc_peering_connection",
                "aws_vpc_endpoint",
                "aws_rds_instance",
                "aws_elb",
                "aws_ec2_subnet",
                "aws_ec2_security_group",
                "aws_ec2_route_table",
                "aws_ec2_network_interface",
                "aws_ec2_network_acl",
                "aws_ec2_nat_gateway",
                "aws_ec2_internet_gateway",
                "aws_alb_target_group",
                "aws_lambda_function",
            ],
        }
    }
    is_default: bool = False

    def delete(self, graph: Graph) -> bool:
        if self.is_default:
            log_msg = f"Not removing the default VPC {self.id} - aborting delete request"
            log.debug(log_msg)
            self.log(log_msg)
            return False

        ec2 = aws_resource(self, "ec2", graph)
        vpc = ec2.Vpc(self.id)
        vpc.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_resource(self, "ec2")
        vpc = ec2.Vpc(self.id)
        vpc.create_tags(Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSVPCQuota(AWSResource, BaseNetworkQuota):
    kind: ClassVar[str] = "aws_vpc_quota"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_vpc"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSS3Bucket(AWSResource, BaseBucket):
    kind: ClassVar[str] = "aws_s3_bucket"

    def delete(self, graph: Graph) -> bool:
        s3 = aws_resource(self, "s3", graph)
        bucket = s3.Bucket(self.name)
        bucket.objects.delete()
        bucket.delete()
        return True

    def set_tags(self, tags: Dict[str, str]) -> bool:
        client = aws_client(self, "s3")
        tag_set = [{"Key": k, "Value": v} for k, v in tags.items()]
        client.put_bucket_tagging(Bucket=self.name, Tagging={"TagSet": tag_set})
        # BaseResource.tags is a setter that will make a copy of the provided
        # dict ref and turn it into a ResourceTagsDict.
        self.tags = tags
        return True

    def update_tag(self, key, value) -> bool:
        # Because S3 tags can not be updated or deleted individually we
        # fetch the current set of tags to keep the race between reading
        # and writing short. Ideally we'd have a revision like in GCP or
        # the ability to modify individual tags instead of all at once.
        tags = self.get_tags()
        tags[key] = value
        return self.set_tags(tags)

    def delete_tag(self, key) -> bool:
        tags = self.get_tags()
        if key in tags:
            del tags[key]
        else:
            raise KeyError(key)
        return self.set_tags(tags)

    def get_tags(self) -> Dict:
        """Fetch the S3 buckets tags from the AWS API."""
        log.debug(f"Fetching tags for {self.rtdname}")
        tags = {}
        client = aws_client(self, "s3")
        try:
            response = client.get_bucket_tagging(Bucket=self.name)
            tags = tags_as_dict(response["TagSet"])
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] != "NoSuchTagSet":
                raise
        return tags

    def refresh_tags(self) -> None:
        """Replace the resources tags with the current tags from the AWS API.

        DO NOT call this method from within update_tag or delete_tag!
        """
        self.tags = self.get_tags()


@define(eq=False, slots=False)
class AWSS3BucketQuota(AWSResource, BaseBucketQuota):
    kind: ClassVar[str] = "aws_s3_bucket_quota"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_s3_bucket"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSELB(AWSResource, BaseLoadBalancer):
    kind: ClassVar[str] = "aws_elb"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_instance"],
            "delete": [],
        }
    }

    def delete(self, graph: Graph) -> bool:
        client = aws_client(self, "elb", graph)
        _ = client.delete_load_balancer(LoadBalancerName=self.name)
        # todo: parse result
        return True

    def update_tag(self, key, value) -> bool:
        client = aws_client(self, "elb")
        client.add_tags(LoadBalancerNames=[self.name], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        client = aws_client(self, "elb")
        client.remove_tags(LoadBalancerNames=[self.name], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSALB(AWSResource, BaseLoadBalancer):
    kind: ClassVar[str] = "aws_alb"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_iam_server_certificate", "aws_alb_target_group"],
            "delete": [],
        }
    }

    def delete(self, graph: Graph) -> bool:
        client = aws_client(self, "elbv2", graph)
        _ = client.delete_load_balancer(LoadBalancerArn=self.arn)
        # todo: block until loadbalancer is gone
        return True

    def update_tag(self, key, value) -> bool:
        client = aws_client(self, "elbv2")
        client.add_tags(ResourceArns=[self.arn], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        client = aws_client(self, "elbv2")
        client.remove_tags(ResourceArns=[self.arn], TagKeys=[key])
        return True


@define(eq=False, slots=False)
class AWSALBTargetGroup(AWSResource, BaseResource):
    kind: ClassVar[str] = "aws_alb_target_group"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_instance"],
            "delete": ["aws_alb"],
        }
    }

    target_type: str = ""

    def delete(self, graph: Graph) -> bool:
        client = aws_client(self, "elbv2", graph)
        _ = client.delete_target_group(TargetGroupArn=self.arn)
        # todo: parse result
        return True

    def update_tag(self, key, value) -> bool:
        client = aws_client(self, "elbv2")
        client.add_tags(ResourceArns=[self.arn], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        client = aws_client(self, "elbv2")
        client.remove_tags(ResourceArns=[self.arn], TagKeys=[key])
        return True


@define(eq=False, slots=False)
class AWSELBQuota(AWSResource, BaseLoadBalancerQuota):
    kind: ClassVar[str] = "aws_elb_quota"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_elb"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSALBQuota(AWSResource, BaseLoadBalancerQuota):
    kind: ClassVar[str] = "aws_alb_quota"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_alb"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSEC2InternetGateway(AWSResource, BaseGateway):
    kind: ClassVar[str] = "aws_ec2_internet_gateway"

    def pre_delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        internet_gateway = ec2.InternetGateway(self.id)
        for predecessor in self.predecessors(graph, edge_type=EdgeType.delete):
            if isinstance(predecessor, AWSVPC):
                log_msg = f"Detaching {predecessor.kind} {predecessor.dname}"
                self.log(log_msg)
                log.debug(f"{log_msg} for deletion of {self.kind} {self.dname}")
                internet_gateway.detach_from_vpc(VpcId=predecessor.id)
        return True

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        internet_gateway = ec2.InternetGateway(self.id)
        internet_gateway.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2NATGateway(AWSResource, BaseGateway):
    kind: ClassVar[str] = "aws_ec2_nat_gateway"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_network_interface"],
            "delete": [],
        }
    }

    nat_gateway_status: str = ""

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_client(self, "ec2", graph)
        ec2.delete_nat_gateway(NatGatewayId=self.id)
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2InternetGatewayQuota(AWSResource, BaseGatewayQuota):
    kind: ClassVar[str] = "aws_ec2_internet_gateway_quota"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_internet_gateway"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSEC2SecurityGroup(AWSResource, BaseSecurityGroup):
    kind: ClassVar[str] = "aws_ec2_security_group"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [
                "aws_vpc_endpoint",
                "aws_rds_instance",
                "aws_elb",
                "aws_ec2_network_interface",
                "aws_lambda_function",
            ],
            "delete": [
                "aws_vpc_endpoint",
                "aws_rds_instance",
                "aws_lambda_function",
            ],
        }
    }

    def pre_delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        security_group = ec2.SecurityGroup(self.id)
        remove_ingress = []
        remove_egress = []

        for permission in security_group.ip_permissions:
            if "UserIdGroupPairs" in permission and len(permission["UserIdGroupPairs"]) > 0:
                p = copy.deepcopy(permission)
                remove_ingress.append(p)
                log.debug(f"Adding incoming permission {p} of {self.kind} {self.dname} to removal list")

        for permission in security_group.ip_permissions_egress:
            if "UserIdGroupPairs" in permission and len(permission["UserIdGroupPairs"]) > 0:
                p = copy.deepcopy(permission)
                remove_egress.append(p)
                log.debug(f"Adding outgoing permission {p} of {self.kind} {self.dname} to removal list")

        if len(remove_ingress) > 0:
            security_group.revoke_ingress(IpPermissions=remove_ingress)

        if len(remove_egress) > 0:
            security_group.revoke_egress(IpPermissions=remove_egress)

        return True

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        security_group = ec2.SecurityGroup(self.id)
        security_group.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2RouteTable(AWSResource, BaseRoutingTable):
    kind: ClassVar[str] = "aws_ec2_route_table"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_vpc_endpoint"],
            "delete": ["aws_vpc_endpoint"],
        }
    }

    def pre_delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        rt = ec2.RouteTable(self.id)
        for rta in rt.associations:
            if not rta.main:
                log_msg = f"Deleting route table association {rta.id}"
                self.log(log_msg)
                log.debug(f"{log_msg} for cleanup of {self.kind} {self.dname}")
                rta.delete()
        return True

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        rt = ec2.RouteTable(self.id)
        rt.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSVPCPeeringConnection(AWSResource, BasePeeringConnection):
    kind: ClassVar[str] = "aws_vpc_peering_connection"
    vpc_peering_connection_status: str = ""

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_client(self, "ec2", graph)
        ec2.delete_vpc_peering_connection(VpcPeeringConnectionId=self.id)
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSVPCEndpoint(AWSResource, BaseEndpoint):
    kind: ClassVar[str] = "aws_vpc_endpoint"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_network_interface"],
            "delete": [],
        }
    }
    vpc_endpoint_type: str = ""
    vpc_endpoint_status: str = ""

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_client(self, "ec2", graph)
        ec2.delete_vpc_endpoints(VpcEndpointIds=[self.id])
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2NetworkAcl(AWSResource, BaseNetworkAcl):
    kind: ClassVar[str] = "aws_ec2_network_acl"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [],
            "delete": ["aws_ec2_subnet"],
        }
    }
    is_default: bool = False

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_client(self, "ec2", graph)
        ec2.delete_network_acl(NetworkAclId=self.id)
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSEC2NetworkInterface(AWSResource, BaseNetworkInterface):
    kind: ClassVar[str] = "aws_ec2_network_interface"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_elastic_ip"],
            "delete": ["aws_vpc_endpoint", "aws_ec2_nat_gateway", "aws_ec2_instance"],
        }
    }

    def delete(self, graph: Graph) -> bool:
        ec2 = aws_resource(self, "ec2", graph)
        network_interface = ec2.NetworkInterface(self.id)
        network_interface.delete()
        return True

    def update_tag(self, key, value) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.create_tags(Resources=[self.id], Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        ec2 = aws_client(self, "ec2")
        ec2.delete_tags(Resources=[self.id], Tags=[{"Key": key}])
        return True


@define(eq=False, slots=False)
class AWSRDSInstance(AWSResource, BaseDatabase):
    kind: ClassVar[str] = "aws_rds_instance"
    volume_kms_key_id: Optional[str] = None


@define(eq=False, slots=False)
class AWSIAMUser(AWSResource, BaseUser):
    kind: ClassVar[str] = "aws_iam_user"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_iam_access_key"],
            "delete": ["aws_iam_policy"],
        }
    }
    user_policies: List = field(factory=list)

    def pre_delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        user = iam.User(self.name)
        for successor in self.successors(graph, edge_type=EdgeType.delete):
            if isinstance(successor, AWSIAMPolicy):
                log_msg = f"Detaching {successor.rtdname}"
                self.log(log_msg)
                log.debug(f"{log_msg} for deletion of {self.rtdname}")
                user.detach_policy(PolicyArn=successor.arn)

        iam = aws_client(self, "iam", graph)
        for user_policy in self.user_policies:
            log_msg = f"Deleting inline policy {user_policy}"
            self.log(log_msg)
            log.debug(f"{log_msg} for deletion of {self.rtdname}")
            iam.delete_user_policy(
                UserName=self.name,
                PolicyName=user_policy,
            )
        return True

    def delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        user = iam.User(self.name)
        user.delete()
        return True


@define(eq=False, slots=False)
class AWSIAMGroup(AWSResource, BaseGroup):
    kind: ClassVar[str] = "aws_iam_group"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_iam_user"],
            "delete": ["aws_iam_policy"],
        }
    }
    group_policies: List = field(factory=list)

    def pre_delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        group = iam.Group(self.name)
        for successor in self.successors(graph, edge_type=EdgeType.delete):
            if isinstance(successor, AWSIAMPolicy):
                log_msg = f"Detaching {successor.rtdname}"
                self.log(log_msg)
                log.debug(f"{log_msg} for deletion of {self.rtdname}")
                group.detach_policy(PolicyArn=successor.arn)

        iam = aws_client(self, "iam", graph)
        for group_policy in self.group_policies:
            log_msg = f"Deleting inline policy {group_policy}"
            self.log(log_msg)
            log.debug(f"{log_msg} for deletion of {self.rtdname}")
            iam.delete_group_policy(
                GroupName=self.name,
                PolicyName=group_policy,
            )
        return True

    def delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        group = iam.Group(self.name)
        group.delete()
        return True


@define(eq=False, slots=False)
class AWSIAMRole(AWSResource, BaseRole):
    kind: ClassVar[str] = "aws_iam_role"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_iam_policy", "aws_iam_instance_profile", "aws_eks_cluster", "aws_lambda_function"],
            "delete": ["aws_iam_policy", "aws_iam_instance_profile", "aws_eks_cluster", "aws_lambda_function"],
        }
    }

    role_policies: List = field(factory=list)

    def pre_delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        role = iam.Role(self.name)
        for successor in self.successors(graph, edge_type=EdgeType.delete):
            if isinstance(successor, AWSIAMPolicy):
                log_msg = f"Detaching {successor.rtdname}"
                self.log(log_msg)
                log.debug(f"{log_msg} for deletion of {self.rtdname}")
                role.detach_policy(PolicyArn=successor.arn)

        iam = aws_client(self, "iam", graph)
        for role_policy in self.role_policies:
            log_msg = f"Deleting inline policy {role_policy}"
            self.log(log_msg)
            log.debug(f"{log_msg} for deletion of {self.rtdname}")
            iam.delete_role_policy(
                RoleName=self.name,
                PolicyName=role_policy,
            )
        return True

    def delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        role = iam.Role(self.name)
        role.delete()
        return True


@define(eq=False, slots=False)
class AWSIAMPolicy(AWSResource, BasePolicy):
    kind: ClassVar[str] = "aws_iam_policy"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_iam_user", "aws_iam_group"],
            "delete": [],
        }
    }

    def delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        policy = iam.Policy(self.arn)
        policy.delete()
        return True


@define(eq=False, slots=False)
class AWSIAMInstanceProfile(AWSResource, BaseInstanceProfile):
    kind: ClassVar[str] = "aws_iam_instance_profile"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_instance"],
            "delete": [],
        }
    }

    def pre_delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        instance_profile = iam.InstanceProfile(self.name)
        for predecessor in self.predecessors(graph, edge_type=EdgeType.delete):
            if isinstance(predecessor, AWSIAMRole):
                log_msg = f"Detaching {predecessor.rtdname}"
                self.log(log_msg)
                log.debug(f"{log_msg} for deletion of {self.rtdname}")
                instance_profile.remove_role(RoleName=predecessor.name)
        return True

    def delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        instance_profile = iam.InstanceProfile(self.name)
        instance_profile.delete()
        return True


@define(eq=False, slots=False)
class AWSIAMAccessKey(AWSResource, BaseAccessKey):
    kind: ClassVar[str] = "aws_iam_access_key"

    user_name: Optional[str] = None
    access_key_last_used_region: Optional[str] = None
    access_key_last_used_service_name: Optional[str] = None

    def delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        access_key = iam.AccessKey(self.user_name, self.id)
        access_key.delete()
        return True


@define(eq=False, slots=False)
class AWSIAMServerCertificate(AWSResource, BaseCertificate):
    kind: ClassVar[str] = "aws_iam_server_certificate"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [],
            "delete": ["aws_alb"],
        }
    }
    path: str = None

    def delete(self, graph: Graph) -> bool:
        iam = aws_resource(self, "iam", graph)
        certificate = iam.ServerCertificate(self.name)
        certificate.delete()
        return True


@define(eq=False, slots=False)
class AWSIAMServerCertificateQuota(AWSResource, BaseCertificateQuota):
    kind: ClassVar[str] = "aws_iam_server_certificate_quota"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_iam_server_certificate"],
            "delete": [],
        }
    }


@define(eq=False, slots=False)
class AWSCloudFormationStack(AWSResource, BaseStack):
    kind: ClassVar[str] = "aws_cloudformation_stack"

    def delete(self, graph: Graph) -> bool:
        cf = aws_resource(self, "cloudformation", graph)
        stack = cf.Stack(self.name)
        stack.delete()
        return True

    class ModificationMode(Enum):
        """Defines Tag modification mode"""

        UPDATE = auto()
        DELETE = auto()

    def update_tag(self, key, value) -> bool:
        return self._modify_tag(key, value, mode=AWSCloudFormationStack.ModificationMode.UPDATE)

    def delete_tag(self, key) -> bool:
        return self._modify_tag(key, mode=AWSCloudFormationStack.ModificationMode.DELETE)

    def _modify_tag(self, key, value=None, mode=None, wait=False) -> bool:
        tags = dict(self.tags)
        if mode == AWSCloudFormationStack.ModificationMode.DELETE:
            if not self.tags.get(key):
                raise KeyError(key)
            del tags[key]
        elif mode == AWSCloudFormationStack.ModificationMode.UPDATE:
            if self.tags.get(key) == value:
                return True
            tags.update({key: value})
        else:
            return False

        cf = aws_resource(self, "cloudformation")
        stack = cf.Stack(self.name)
        stack = self.wait_for_completion(stack, cf)
        response = stack.update(
            Capabilities=["CAPABILITY_NAMED_IAM"],
            UsePreviousTemplate=True,
            Tags=[{"Key": label, "Value": value} for label, value in tags.items()],
            Parameters=[
                {"ParameterKey": parameter, "UsePreviousValue": True} for parameter in self.stack_parameters.keys()
            ],
        )
        if response.get("ResponseMetadata", {}).get("HTTPStatusCode", 0) != 200:
            raise RuntimeError(f"Error updating AWS Cloudformation Stack {self.dname} for {mode.name} of tag {key}")
        if wait:
            self.wait_for_completion(stack, cf)
        self.tags = tags
        return True

    def wait_for_completion(self, stack, cloudformation_resource, timeout=300):
        start_utime = time.time()
        while stack.stack_status.endswith("_IN_PROGRESS"):
            if time.time() > start_utime + timeout:
                raise TimeoutError(
                    (
                        f"AWS Cloudformation Stack {self.dname} tag update timed out "
                        f"after {timeout} seconds with status {stack.stack_status}"
                    )
                )
            time.sleep(5)
            stack = cloudformation_resource.Stack(stack.name)
        return stack


@define(eq=False, slots=False)
class AWSEKSCluster(AWSResource, BaseResource):
    kind: ClassVar[str] = "aws_eks_cluster"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_eks_nodegroup"],
            "delete": ["aws_eks_nodegroup"],
        }
    }
    cluster_status: str = ""
    cluster_endpoint: str = ""

    def delete(self, graph: Graph) -> bool:
        eks = aws_client(self, "eks", graph)
        eks.delete_cluster(name=self.name)
        return True

    def update_tag(self, key, value) -> bool:
        eks = aws_client(self, "eks")
        eks.tag_resource(resourceArn=self.arn, tags={key: value})
        return True

    def delete_tag(self, key) -> bool:
        eks = aws_client(self, "eks")
        eks.untag_resource(resourceArn=self.arn, tagKeys=[key])
        return True


@define(eq=False, slots=False)
class AWSEKSNodegroup(AWSResource, BaseResource):
    kind: ClassVar[str] = "aws_eks_nodegroup"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_autoscaling_group"],
            "delete": [],
        }
    }

    cluster_name: str = ""
    nodegroup_status: str = ""

    def delete(self, graph: Graph) -> bool:
        eks = aws_client(self, "eks", graph)
        eks.delete_nodegroup(clusterName=self.cluster_name, nodegroupName=self.name)
        return True

    def update_tag(self, key, value) -> bool:
        eks = aws_client(self, "eks")
        eks.tag_resource(resourceArn=self.arn, tags={key: value})
        return True

    def delete_tag(self, key) -> bool:
        eks = aws_client(self, "eks")
        eks.untag_resource(resourceArn=self.arn, tagKeys=[key])
        return True


@define(eq=False, slots=False)
class AWSAutoScalingGroup(AWSResource, BaseAutoScalingGroup):
    kind: ClassVar[str] = "aws_autoscaling_group"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_ec2_instance"],
            "delete": ["aws_eks_nodegroup"],
        }
    }

    def delete(self, graph: Graph, force_delete: bool = True) -> bool:
        client = aws_client(self, "autoscaling", graph)
        client.delete_auto_scaling_group(AutoScalingGroupName=self.name, ForceDelete=force_delete)
        return True

    def update_tag(self, key, value) -> bool:
        client = aws_client(self, "autoscaling")
        client.create_or_update_tags(
            Tags=[
                {
                    "ResourceId": self.name,
                    "ResourceType": "auto-scaling-group",
                    "Key": key,
                    "Value": value,
                    "PropagateAtLaunch": True,
                }
            ]
        )
        return True

    def delete_tag(self, key) -> bool:
        client = aws_client(self, "autoscaling")
        client.delete_tags(
            Tags=[
                {
                    "ResourceId": self.name,
                    "ResourceType": "auto-scaling-group",
                    "Key": key,
                    "Value": self.tags[key],
                    "PropagateAtLaunch": True,
                }
            ]
        )
        return True


@define(eq=False, slots=False)
class AWSCloudwatchAlarm(AWSResource, BaseResource):
    kind: ClassVar[str] = "aws_cloudwatch_alarm"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [],
            "delete": [],
        }
    }

    actions_enabled: bool = False
    alarm_description: Optional[str] = None
    alarm_actions: Optional[List] = field(factory=list)
    comparison_operator: Optional[str] = None
    dimensions: Optional[List] = field(factory=list)
    evaluation_periods: Optional[int] = 0
    insufficient_data_actions: Optional[List] = field(factory=list)
    metric_name: Optional[str] = None
    namespace: Optional[str] = None
    ok_actions: Optional[List] = field(factory=list)
    period: Optional[int] = 0
    state_value: Optional[str] = None
    statistic: Optional[str] = None
    threshold: Optional[float] = 0.0

    def delete(self, graph: Graph) -> bool:
        cloudwatch = aws_resource(self, "cloudwatch", graph)
        alarm = cloudwatch.Alarm(self.name)
        alarm.delete()
        return True

    def update_tag(self, key, value) -> bool:
        client = aws_client(self, "cloudwatch")
        client.tag_resource(ResourceARN=self.arn, Tags=[{"Key": key, "Value": value}])
        return True

    def delete_tag(self, key) -> bool:
        client = aws_client(self, "cloudwatch")
        client.untag_resource(ResourceARN=self.arn, TagKeys=[key])
        return True


@define(eq=False, slots=False)
class AWSCloudFormationStackSet(AWSResource, BaseResource):
    kind: ClassVar[str] = "aws_cloudformation_stack_set"
    description: Optional[str] = None
    stack_set_status: Optional[str] = None
    stack_set_parameters: Dict = field(factory=dict)
    stack_set_capabilities: Optional[List[str]] = field(factory=list)
    stack_set_administration_role_arn: Optional[str] = None
    stack_set_execution_role_name: Optional[str] = None
    stack_set_drift_detection_details: Optional[Dict[str, Any]] = field(factory=dict)
    stack_set_last_drift_check_timestamp: Optional[datetime] = None
    stack_set_auto_deployment: Optional[Dict[str, bool]] = field(factory=dict)
    stack_set_permission_model: Optional[str] = None
    stack_set_organizational_unit_ids: Optional[List[str]] = field(factory=list)
    stack_set_managed_execution_active: Optional[bool] = None

    def delete(self, graph: Graph) -> bool:
        cf = aws_client(self, "cloudformation", graph)
        cf.delete_stack_set(StackSetName=self.name)
        return True

    class ModificationMode(Enum):
        """Defines Tag modification mode"""

        UPDATE = auto()
        DELETE = auto()

    def update_tag(self, key, value) -> bool:
        return self._modify_tag(key, value, mode=AWSCloudFormationStackSet.ModificationMode.UPDATE)

    def delete_tag(self, key) -> bool:
        return self._modify_tag(key, mode=AWSCloudFormationStackSet.ModificationMode.DELETE)

    def _modify_tag(self, key, value=None, mode=None) -> bool:
        tags = dict(self.tags)
        if mode == AWSCloudFormationStackSet.ModificationMode.DELETE:
            if not self.tags.get(key):
                raise KeyError(key)
            del tags[key]
        elif mode == AWSCloudFormationStackSet.ModificationMode.UPDATE:
            if self.tags.get(key) == value:
                return True
            tags.update({key: value})
        else:
            return False

        cf = aws_client(self, "cloudformation")
        response = cf.update_stack_set(
            StackSetName=self.name,
            Capabilities=["CAPABILITY_NAMED_IAM"],
            UsePreviousTemplate=True,
            Tags=[{"Key": label, "Value": value} for label, value in tags.items()],
            Parameters=[
                {"ParameterKey": parameter, "UsePreviousValue": True} for parameter in self.stack_set_parameters.keys()
            ],
        )
        if response.get("ResponseMetadata", {}).get("HTTPStatusCode", 0) != 200:
            raise RuntimeError(
                "Error updating AWS Cloudformation Stack Set" f" {self.dname} for {mode.name} of tag {key}"
            )
        self.tags = tags
        return True


@define(eq=False, slots=False)
class AWSRoute53Zone(AWSResource, BaseDNSZone):
    kind: ClassVar[str] = "aws_route53_zone"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_route53_resource_record_set"],
            "delete": [],
        }
    }

    zone_caller_reference: Optional[str] = None
    zone_config: Optional[Dict[str, Union[str, bool]]] = None
    zone_resource_record_set_count: Optional[int] = None
    zone_linked_service: Optional[Dict[str, str]] = None


@define(eq=False, slots=False)
class AWSRoute53ResourceRecordSet(AWSResource, BaseDNSRecordSet):
    kind: ClassVar[str] = "aws_route53_resource_record_set"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": ["aws_route53_resource_record"],
            "delete": ["aws_route53_resource_record"],
        }
    }
    record_set_identifier: Optional[str] = None
    record_region: Optional[str] = None
    record_geo_location: Optional[Dict[str, str]] = None
    record_failover: Optional[str] = None
    record_multi_value_answer: Optional[bool] = None
    record_alias_target: Optional[Dict[str, Union[str, bool]]] = None
    record_health_check_id: Optional[str] = None
    record_traffic_policy_instance_id: Optional[str] = None


@define(eq=False, slots=False)
class AWSRoute53ResourceRecord(AWSResource, BaseDNSRecord):
    kind: ClassVar[str] = "aws_route53_resource_record"
    reference_kinds: ClassVar[ModelReference] = {
        "successors": {
            "default": [],
            "delete": [],
        }
    }
